//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#include "ez_bmp.hpp"
#include "ez_uail_gui.hpp"
#include "ez_wave.hpp"

//############################################################################
ez_uail_frame::ez_uail_frame(ez_uail_point(*F)(int, int), int points_per_frame, int frame_index)
{
    is_selected         = false            ;
    is_unique           = true             ;
    palette_index       = EZ_EZFB_HUES     ;
    error               = EZ_UAIL_OK       ;
    format              = EZ_UAIL_3D_VECTOR;
    ez_uail_point point = F(0, frame_index);
    point.blank();
    reserve(points_per_frame + 2);    
    push_back(point);
    for(int p = 0; p < points_per_frame; p++)
        push_back(F(p, frame_index));
    push_back(point);
    quantity = (u_short)size();
}

//############################################################################
void ez_uail_frame::save_as_ilda(string file)
{
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".ild"
      )
        file += ".ild";
    ofstream out((EZ_UAIL_FRAME_SHARE + file).c_str(), ios::out | ios::binary);
    if(out.is_open())
    {
        if(palette_index != EZ_ILDA_DEFAULT)
          AIL->to_ofstream_ilda(out);
        else
            to_ofstream_ilda(out);
        out.close();
    }
    return;
}

//############################################################################
void ez_uail_frame::optimize()
{
    minimize                 ();
    add_dwell_points         ();
    add_lit_span_points      ();
    add_blank_span_points    ();
    quantity = (u_short)size ();      
    return;
}

//############################################################################
int ez_uail_frame::from_ifstream_format_3(ifstream& in, const ez_uail_header& header)
{
    int error = EZ_UAIL_OK;
    *this     = header;
    clear();
    //------------------------------------------------------------------------
    if(quantity)
    {
        int             i;
        ez_color        rgb;
        ez_uail_point   point;
        ez_uail_header  next_header;
        //--------------------------------------------------------------------
        reserve(quantity);
        for(i = 0; i < (int)quantity; i++)
        {   
            error |= rgb.from_ifstream_ilda(in);
            if(!error)
                push_back((ez_uail_point)rgb);
            else
                return error;
        }
        //--------------------------------------------------------------------
        next_header.from_ifstream_ilda(in);

        if(quantity != next_header.quantity)
            return EZ_COLOR_POINT_MISSMATCH;

        *this = next_header;

        for(i = 0; i < (int)quantity; i++)
        {   
            error |= point.from_ifstream_ilda(in, next_header.format);
            if(!error)
            {
                at(i).x = point.x;
                at(i).y = point.y;
                at(i).z = point.z;
                at(i).k = point.k;
                at(i).c = point.c;
            }
            else
                return error;
        }
        //--------------------------------------------------------------------
        front().blank();
        //--------------------------------------------------------------------
        palette_index = EZ_TRUE_COLOR;
        sync_rgb_and_palette();
        //--------------------------------------------------------------------
    }
    //------------------------------------------------------------------------
    return error;
}

//############################################################################
int ez_uail_frame::from_ifstream_ilda(ifstream& in, const ez_uail_header& header)
{
    int            i,
                   error = EZ_UAIL_OK;
    ez_uail_point  point;
    //------------------------------------------------------------------------
    *this = header;
    if(    format == EZ_UAIL_3D_COLOR_VECTOR
        || format == EZ_UAIL_2D_COLOR_VECTOR
      )
    {
        palette_index = EZ_TRUE_COLOR;
    }   
    //------------------------------------------------------------------------
    clear();
    if(quantity)
    {
        reserve(quantity);
        for(i = 0; i < (int)quantity; i++)
        {   
            error |= point.from_ifstream_ilda(in, format);
            if(!error)
                push_back(point);
            else
                return error;
        }
        //--------------------------------------------------------------------
        front().blank();
    }
    //------------------------------------------------------------------------
    return error;
}

//############################################################################
int ez_uail_frame::from_ifstream_dxf(ifstream& in, const ez_uail_header& header)
{
    int             i             ;
    float           real_size  = 0,
                    real_scale = 1;
    ez_3D_float     real_min      ,
                    real_max      ,
                    real_offset   ;
    ez_uail_point   point         ;
    ez_real_segment real_points   ;
    //------------------------------------------------------------------------
    real_points.from_ifstream_dxf(in);
    //------------------------------------------------------------------------
    if(real_points.size() > 2) // more that the default original vector
    {
        if(AIL->auto_scale_dxf)
        {
            real_min =  FLT_MAX;
            real_max = -FLT_MAX;
            for(i = 2; i < (int)real_points.size(); i++)
            {
                if(real_points[i].x > real_max.x)    real_max.x = real_points[i].x;
                if(real_points[i].x < real_min.x)    real_min.x = real_points[i].x;
                if(real_points[i].y > real_max.y)    real_max.y = real_points[i].y;
                if(real_points[i].y < real_min.y)    real_min.y = real_points[i].y;
                if(real_points[i].z > real_max.z)    real_max.z = real_points[i].z;
                if(real_points[i].z < real_min.z)    real_min.z = real_points[i].z;
            }
            if(AIL->maintain_dxf_origin)
            {
                if(fabs(real_max.x) > real_size)    real_size = fabs(real_max.x);
                if(fabs(real_min.x) > real_size)    real_size = fabs(real_min.x);
                if(fabs(real_max.y) > real_size)    real_size = fabs(real_max.y);
                if(fabs(real_min.y) > real_size)    real_size = fabs(real_min.y);
                if(fabs(real_max.z) > real_size)    real_size = fabs(real_max.z);
                if(fabs(real_min.z) > real_size)    real_size = fabs(real_min.z);
    
                real_size *= 2;
            }
            else // find the new center of the universe
            {
                real_offset.x = (real_max.x - ((real_max.x - real_min.x) / 2));
                real_offset.y = (real_max.y - ((real_max.y - real_min.y) / 2));
                real_offset.z = (real_max.z - ((real_max.z - real_min.z) / 2));
    
                if(fabs(real_max.x - real_min.x) > real_size)    real_size = fabs(real_max.x - real_min.x);
                if(fabs(real_max.y - real_min.y) > real_size)    real_size = fabs(real_max.y - real_min.y);
                if(fabs(real_max.z - real_min.z) > real_size)    real_size = fabs(real_max.z - real_min.z);
            }
            real_scale = 65000.0 / real_size;
        } // end if(auto_scale_dxf)
        //--------------------------------------------------------------------
        reserve(real_points.size() - 2);
        for(i = 2; i < (int)real_points.size(); i++)
        {
            point.x = (short)((real_points[i].x - real_offset.x) * real_scale);
            point.y = (short)((real_points[i].y - real_offset.y) * real_scale);
            point.z = (short)((real_points[i].z - real_offset.z) * real_scale);
            point.c = real_points[i].c;
            if(real_points[i].is_blank())
                point.blank();
            else
                point.unblank();
            //point.set_unique();
            push_back(point);
        }
        set_rgb_from_palette();
    }
    else
    {
        reserve(2);
        push_back(point);        
        push_back(point);        
    }
    //------------------------------------------------------------------------    
    return EZ_UAIL_OK;
}
/*
//############################################################################
int ez_uail_frame::from_ifstream_shp(ifstream& in, const ez_uail_header& header)
{ 
    get_shp_number(in);
    get_shp_byte_count(in);
    get_shp_name(in);
    get_shp_op_code(in);
    return error;
}  
*/
//############################################################################
int ez_uail_frame::move_selection(ez_3D_float d)
{
    if(number_of_points() > 1)
    {
        ez_uail_frame   frame(*this);
        ez_3D_float     f;
        int             i,
                        out_of_bounds = EZ_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                out_of_bounds = ez_uail_bounds_check(f + d, EZ_UAIL_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                at(i) = ez_uail_point( f + d,
                                       (ez_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return EZ_IN_BOUNDS;
}
//############################################################################
int ez_uail_frame::scale_selection(ez_3D_float m)
{
    if(number_of_points() > 1)
    {
        ez_3D_float     f,
                        center = selected_segment().find_center();

        int             i,
                        out_of_bounds = EZ_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                out_of_bounds = ez_uail_bounds_check(((f - center) * m) + center, EZ_UAIL_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                at(i) = ez_uail_point( ((f - center) * m) + center,
                                       (ez_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_uail_frame::scale_selection_around_point(ez_3D_float m, ez_3D_float p)
{
    if(number_of_points() > 1)
    {
        ez_3D_float     f;

        int             i,
                        out_of_bounds = EZ_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                out_of_bounds = ez_uail_bounds_check(((f - p) * m) + p, EZ_UAIL_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                at(i) = ez_uail_point( ((f - p) * m) + p,
                                       (ez_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_uail_frame::scale_selection_around_fulcrum(ez_3D_float m)
{
    if(number_of_points() > 1)
    {
        ez_3D_float     f;

        int             i,
                        out_of_bounds = EZ_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                out_of_bounds = ez_uail_bounds_check(((f - AIL->fulcrum) * m) + AIL->fulcrum, EZ_UAIL_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                at(i) = ez_uail_point( ((f - AIL->fulcrum) * m) + AIL->fulcrum,
                                       (ez_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_uail_frame::rotate_selection(ez_3D_float a)
{
    if(number_of_points() > 1)
    {
        ez_3D_float     f,
                        center = selected_segment().find_center();

        int             i,
                        out_of_bounds = EZ_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                out_of_bounds = ez_uail_bounds_check(rotate_point_around_point(f, center, a), EZ_UAIL_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                at(i) = ez_uail_point( rotate_point_around_point(f, center, a),
                                       (ez_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_uail_frame::rotate_selection_around_point(ez_3D_float a, ez_3D_float p)
{
    if(number_of_points() > 1)
    {
        ez_3D_float     f;

        int             i,
                        out_of_bounds = EZ_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                out_of_bounds = ez_uail_bounds_check(rotate_point_around_point(f, p, a), EZ_UAIL_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                at(i) = ez_uail_point( rotate_point_around_point(f, p, a),
                                       (ez_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_uail_frame::rotate_selection_around_fulcrum(ez_3D_float a)
{
    if(number_of_points() > 1)
    {
        ez_3D_float     f;

        int             i,
                        out_of_bounds = EZ_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                out_of_bounds = ez_uail_bounds_check(rotate_point_around_point(f, AIL->fulcrum, a), EZ_UAIL_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                at(i) = ez_uail_point( rotate_point_around_point(f, AIL->fulcrum, a),
                                       (ez_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
void ez_uail_frame::next_segment_select()
{
    int i, start, end;
    find_segment(egg, start, end, i);
    find_segment(i + 1, egg, spider);
    return;
}

//############################################################################
void ez_uail_frame::next_segment_egg()
{
    int i, start, end;
    find_segment(egg, start, end, i);
    find_segment(i + 1, egg, end);
    return;
}

//############################################################################
void ez_uail_frame::next_segment_spider()
{
    int i, start, end;
    find_segment(spider, start, end, i);
    find_segment(i + 1, start, spider);
    return;
}

//############################################################################
void ez_uail_frame::previous_segment_select()
{
    int i, start, end;
    find_segment(egg, start, end, i);
    find_segment(i - 1, egg, spider);
    return;
}

//############################################################################
void ez_uail_frame::previous_segment_egg()
{
    int i, start, end;
    find_segment(egg, start, end, i);
    find_segment(i - 1, egg, end);
    return;
}

//############################################################################
void ez_uail_frame::previous_segment_spider()
{
    int i, start, end;
    find_segment(spider, start, end, i);
    find_segment(i - 1, start, spider);
    return;
}

//############################################################################
void ez_uail_frame::recolor_select()
{
    if(quantity > 1)
    {
        int i;
        //--------------------------------------------------------------------                        
        if(palette_index != EZ_TRUE_COLOR)
        {
            for(i = 0; i < number_of_points(); i++)
                if(    at(i).is_color()
                    && is_index_selected(i)
                  )
                {
                    at(i).c = AIL->selected_palette_color;
                    at(i).r = AIL->palette_picker(palette_index)[AIL->selected_palette_color].r;
                    at(i).g = AIL->palette_picker(palette_index)[AIL->selected_palette_color].g;
                    at(i).b = AIL->palette_picker(palette_index)[AIL->selected_palette_color].b;
                }
        }
        //--------------------------------------------------------------------                        
        else
        {
            for(i = 0; i < number_of_points(); i++)
                if(    at(i).is_color()
                    && is_index_selected(i)
                  )
                {
                    at(i).r = AIL->palette_picker(palette_index)[AIL->selected_palette_color].r;
                    at(i).g = AIL->palette_picker(palette_index)[AIL->selected_palette_color].g;
                    at(i).b = AIL->palette_picker(palette_index)[AIL->selected_palette_color].b;
                }            
        }
        //--------------------------------------------------------------------                        
    }
    return;
}

//############################################################################
void ez_uail_frame::recolor_black_in_select()
{
    if(quantity > 1)
    {
        int i;
        //--------------------------------------------------------------------                        
        if(palette_index != EZ_TRUE_COLOR)
        {
            for(i = 0; i < number_of_points(); i++)
                if(is_index_selected(i) && at(i).is_black())
                {
                    at(i).c = AIL->selected_palette_color;
                    at(i).r = AIL->palette_picker(palette_index)[AIL->selected_palette_color].r;
                    at(i).g = AIL->palette_picker(palette_index)[AIL->selected_palette_color].g;
                    at(i).b = AIL->palette_picker(palette_index)[AIL->selected_palette_color].b;
                }
        }
        //--------------------------------------------------------------------                        
        else
        {
            for(i = 0; i < number_of_points(); i++)
                if(is_index_selected(i) && at(i).is_black())
                {
                    at(i).r = AIL->palette_picker(palette_index)[AIL->selected_palette_color].r;
                    at(i).g = AIL->palette_picker(palette_index)[AIL->selected_palette_color].g;
                    at(i).b = AIL->palette_picker(palette_index)[AIL->selected_palette_color].b;
                }            
        }
        //--------------------------------------------------------------------                        
    }
    return;
}

//############################################################################
void ez_uail_frame::blank_points()
{
    if(quantity > 1)
    {
        for(int i = 0; i < number_of_points(); i++)
            if(is_index_selected(i))
                at(i).blank();
    }
    return;
}

//############################################################################
void ez_uail_frame::unblank_points()
{
    if(quantity > 1)
    {
        for(int i = 1; i < number_of_points(); i++)
            if(is_index_selected(i))
                at(i).unblank();
    }
    return;
}

//############################################################################
void ez_uail_frame::black_points()
{
    if(quantity > 1)
    {
        for(int i = 1; i < number_of_points(); i++)
            if(is_index_selected(i) && at(i).is_blank())
            {
                at(i).r = at(i).g = at(i).b = 0x00;
                at(i).c = AIL->palette_picker(palette_index).black;
                at(i).unblank();
            }
    }
    return;
}

//############################################################################
void ez_uail_frame::unblack_points()
{
    if(quantity > 1)
    {
        for(int i = 1; i < number_of_points(); i++)
            if(is_index_selected(i) && at(i).color() == 0x00)
                at(i).blank();
    }
    return;
}

//############################################################################
void ez_uail_frame::blank_equivalent_vectors()
{
    ez_uail_frame_set  segments;
    int                i,
                       j;
    //------------------------------------------------------------------------
    fracture_segments();
    segments = find_segments();
    //------------------------------------------------------------------------
    for(i = 0; i < segments.number_of_frames() - 1; i++)
    {
        for(j = i + 1; j < segments.number_of_frames(); j++)
            if(    segments[j] == segments[i]
                || segments[j] == segments[i].reverse()
              )
                segments[j].blank_all_points();
    }
    //------------------------------------------------------------------------
    *this = segments.sum_of_frames();
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void ez_uail_frame::minimize()
{
    if(!AIL->allow_lit_black)
        convert_black_to_blank();
    //------------------------------------------------------------------------
    remove_dwell_points ();
    reduce_blank_vectors();
    reduce_lit_vectors  ();
    //------------------------------------------------------------------------
    if(AIL->no_equivalent_vectors)
    {
        blank_equivalent_vectors();
        reduce_blank_vectors    ();
    }
    //------------------------------------------------------------------------
    if(AIL->no_dots)
        remove_dots ();
    else
        enhance_dots();
    //------------------------------------------------------------------------
    bond_segments();
    validate     ();
    //------------------------------------------------------------------------
    return;
}

//############################################################################
ez_uail_frame& ez_uail_frame::add_new_point(int point_index)
{
    ez_uail_segment::add_new_point(point_index);
    quantity = (u_short)size();
    if(    quantity > 2
        && point_index >= 0
        && point_index <  quantity
      )
    {
        if((at(point_index) | at(point_index + 1)) > AIL->insignificant_distance)
            spider = point_index + 1;
        else
            spider = point_index;
    }
    else
        egg = spider = point_index = 0;
    return *this;
}

//############################################################################
ez_uail_frame& ez_uail_frame::break_segment()
{
    ez_uail_segment::break_segment(spider);
    quantity = (u_short)size();
    return *this;
}

//############################################################################
ez_uail_frame& ez_uail_frame::connect_the_dots()
{
    ez_uail_segment::connect_the_dots(egg, spider);
    quantity = (u_short)size();
    return *this;
}

//############################################################################
ez_uail_frame& ez_uail_frame::to_dots()
{
    int            i;
    ez_uail_frame  frame(palette_index);
    //--------------------------------------------------------------------
    frame.reserve(number_of_points() + size_of_selection());
    //--------------------------------------------------------------------
    for(i = 0; i < number_of_points(); i++)
    {
        if(    at(i).is_lit()
            && at_index_has_magnitude(i)
          )
        {
            frame += at(i);
            frame.back().r = 0;
            frame.back().g = 0;
            frame.back().b = 0;
            frame.back().c = AIL->palette_picker(palette_index).black;
        }
        frame += at(i);
    }
    //--------------------------------------------------------------------
    *this = frame;
    quantity = (u_short)size();
    //--------------------------------------------------------------------
    return *this;
}

//############################################################################
ez_uail_frame& ez_uail_frame::selected_to_dots()
{
    int            i,
                   new_spider = spider;
    ez_uail_frame  frame(palette_index);
    //--------------------------------------------------------------------
    frame.reserve(number_of_points() + size_of_selection());
    //--------------------------------------------------------------------
    frame.push_back(front());
    for(i = 1; i < number_of_points(); i++)
    {
        if(    is_index_selected(i)            
            && at(i).is_lit()
            && at_index_has_magnitude(i)
          )
        {
            frame += at(i);
            frame.back().r = 0;
            frame.back().g = 0;
            frame.back().b = 0;
            frame.back().c = AIL->palette_picker(palette_index).black;
            new_spider++;
        }
        frame += at(i);
    }
    //--------------------------------------------------------------------
    spider = new_spider;
    *this = frame;
    quantity = (u_short)size();
    //--------------------------------------------------------------------
    return *this;
}

//############################################################################
void ez_uail_frame::to_ofstream_ilda(ofstream& out)
{
    char number[7];
    if(AIL->dumb_file_format)
    {
        sprintf(number, "%06d", identity);
        name = "ES" + string(number);
        ez_uail_header::to_ofstream_ilda(out);
        for(int i = 0; i < (int)quantity; i++)
            at(i).to_ofstream_ilda(out, format);
    }
    else
    {
        if(palette_index == EZ_TRUE_COLOR)
        {
            ez_uail_header  header  = (ez_uail_header)*this;
            ez_palette      palette = as_ez_palette();
            header.identity--;
            sprintf(number, "%06d", header.identity);
            header.format    = EZ_UAIL_COLOR_TABLE;
            header.name      = "ES" + string(number);
            palette.to_ofstream_ilda(out, header);
        }
        else
        {
            sprintf(number, "%06d", identity);
            name = "ES" + string(number);
        }
        ez_uail_header::to_ofstream_ilda(out);
        for(int i = 0; i < (int)quantity; i++)
            at(i).to_ofstream_ilda(out, format);
    }
    return;
}

//############################################################################
void ez_uail_frame::render_in_ezfb_bitmap(string ui_message) const
{
    u_short     i;

    ez_3D_float _0,
                _1;
                
    ez_color    bit_masked;

    u_int       select_start_color,
                select_end_color  ,
                bmp_bg_color = BMP.make_rgb(AIL->bg_color.r, AIL->bg_color.g, AIL->bg_color.b);

    int         xx,
                yy,
                x_off        = BMP.yres + EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor,
                y_off        =   EZ_PALETTE_MAX * AIL->font_size_factor
                               + (EZ_STATS_DISPLAY_OFFSET + 24) * AIL->font_size_factor,
                y_line       = 8 * AIL->font_size_factor,
                bmp_center   = BMP.yres / 2;

    float       scale        = BMP.yres / 65536.0;

    bool        rotate_view  = (AIL->view_angle  != 0.0),
                offset_view  = (AIL->view_offset != 0.0),
                scale_view   = (AIL->view_scale  != 1.0);
    //------------------------------------------------------------------------
    if(AIL->show_background_bitmap)
        bmp_copy(&BMP, &BACKGROUND_BMP);
    else
        bmp_copy(&BMP, &BACKGROUND);
    //------------------------------------------------------------------------
    if(AIL->show_fixed_bounds)
        bmp_put_rectangle(&BMP, 0, 0, BMP.yres - 1, BMP.yres - 1, BMP.gray, 0);
    //------------------------------------------------------------------------
    if(AIL->show_fixed_origin)
    {
        bmp_put_line_fast(&BMP, 0, BMP.yres / 2, BMP.yres - 1, BMP.yres / 2, BMP.gray);
        bmp_put_line_fast(&BMP, BMP.yres / 2, 0, BMP.yres / 2, BMP.yres - 1, BMP.gray);
    }
    //------------------------------------------------------------------------
    if(AIL->show_floating_axis)
    {
        ez_3D_float   x1(EZ_MIN_SHORT, 0, 0)
                    , x2(EZ_MAX_SHORT, 0, 0)
                    , y1(0, EZ_MIN_SHORT, 0)
                    , y2(0, EZ_MAX_SHORT, 0)
                    , z1(0, 0, EZ_MIN_SHORT)
                    , z2(0, 0, EZ_MAX_SHORT);

        if(rotate_view)
        {
            x1 = rotate_point(x1, AIL->view_angle);
            x2 = rotate_point(x2, AIL->view_angle);
            y1 = rotate_point(y1, AIL->view_angle);
            y2 = rotate_point(y2, AIL->view_angle);
            z1 = rotate_point(z1, AIL->view_angle);
            z2 = rotate_point(z2, AIL->view_angle);
        }
        if(offset_view)
        {
            x1 += AIL->view_offset;
            x2 += AIL->view_offset;
            y1 += AIL->view_offset;
            y2 += AIL->view_offset;
        }
        if(scale_view)
        {
            x1 *= AIL->view_scale;
            x2 *= AIL->view_scale;
            y1 *= AIL->view_scale;
            y2 *= AIL->view_scale;
            z1 *= AIL->view_scale;
            z2 *= AIL->view_scale;
        }
        // X axis
        bmp_put_line_fast(&BMP,             (int)(x1.x * scale) + bmp_center
                              , BMP.yres - ((int)(x1.y * scale) + bmp_center)
                              ,             (int)(x2.x * scale) + bmp_center
                              , BMP.yres - ((int)(x2.y * scale) + bmp_center)
                              , BMP.gray
                         );
        // X axis icon
        bmp_put_line_fast(&BMP,             (int)((x2.x - 500) * scale) + bmp_center
                              , BMP.yres - ((int)((x2.y - 500) * scale) + bmp_center)
                              ,             (int)((x2.x + 500) * scale) + bmp_center
                              , BMP.yres - ((int)((x2.y + 500) * scale) + bmp_center)
                              , BMP.make_rgb(255,0,0)
                         );
        bmp_put_line_fast(&BMP,             (int)((x2.x - 500) * scale) + bmp_center
                              , BMP.yres - ((int)((x2.y + 500) * scale) + bmp_center)
                              ,             (int)((x2.x + 500) * scale) + bmp_center
                              , BMP.yres - ((int)((x2.y - 500) * scale) + bmp_center)
                              , BMP.make_rgb(255,0,0)
                         );
        // Y axis
        bmp_put_line_fast(&BMP,             (int)(y1.x * scale) + bmp_center
                              , BMP.yres - ((int)(y1.y * scale) + bmp_center)
                              ,             (int)(y2.x * scale) + bmp_center
                              , BMP.yres - ((int)(y2.y * scale) + bmp_center)
                              , BMP.gray
                         );
        // Y axis icon
        bmp_put_line_fast(&BMP,             (int)((y2.x - 500) * scale) + bmp_center
                              , BMP.yres - ((int)((y2.y + 500) * scale) + bmp_center)
                              ,             (int)((y2.x      ) * scale) + bmp_center
                              , BMP.yres - ((int)((y2.y      ) * scale) + bmp_center)
                              , BMP.make_rgb(0,255,0)
                         );
        bmp_put_line_fast(&BMP,             (int)((y2.x + 500) * scale) + bmp_center
                              , BMP.yres - ((int)((y2.y + 500) * scale) + bmp_center)
                              ,             (int)((y2.x      ) * scale) + bmp_center
                              , BMP.yres - ((int)((y2.y      ) * scale) + bmp_center)
                              , BMP.make_rgb(0,255,0)
                         );
        bmp_put_line_fast(&BMP,             (int)((y2.x      ) * scale) + bmp_center
                              , BMP.yres - ((int)((y2.y      ) * scale) + bmp_center)
                              ,             (int)((y2.x      ) * scale) + bmp_center
                              , BMP.yres - ((int)((y2.y - 500) * scale) + bmp_center)
                              , BMP.make_rgb(0,255,0)
                         );
        // Z axis
        bmp_put_line_fast(&BMP,             (int)(z1.x * scale) + bmp_center
                              , BMP.yres - ((int)(z1.y * scale) + bmp_center)
                              ,             (int)(z2.x * scale) + bmp_center
                              , BMP.yres - ((int)(z2.y * scale) + bmp_center)
                              , BMP.gray
                         );
        // Z axis icon
        bmp_put_line_fast(&BMP,             (int)((z2.x - 500) * scale) + bmp_center
                              , BMP.yres - ((int)((z2.y + 500) * scale) + bmp_center)
                              ,             (int)((z2.x + 500) * scale) + bmp_center
                              , BMP.yres - ((int)((z2.y + 500) * scale) + bmp_center)
                              , BMP.make_rgb(0,0,255)
                         );
        bmp_put_line_fast(&BMP,             (int)((z2.x + 500) * scale) + bmp_center
                              , BMP.yres - ((int)((z2.y + 500) * scale) + bmp_center)
                              ,             (int)((z2.x - 500) * scale) + bmp_center
                              , BMP.yres - ((int)((z2.y - 500) * scale) + bmp_center)
                              , BMP.make_rgb(0,0,255)
                         );
        bmp_put_line_fast(&BMP,             (int)((z2.x - 500) * scale) + bmp_center
                              , BMP.yres - ((int)((z2.y - 500) * scale) + bmp_center)
                              ,             (int)((z2.x + 500) * scale) + bmp_center
                              , BMP.yres - ((int)((z2.y - 500) * scale) + bmp_center)
                              , BMP.make_rgb(0,0,255)
                         );
    }
    //------------------------------------------------------------------------
    if(AIL->show_floating_bounds)
    {
        ez_3D_float   a(EZ_MIN_SHORT, EZ_MIN_SHORT, EZ_MIN_SHORT)
                    , b(EZ_MAX_SHORT, EZ_MIN_SHORT, EZ_MIN_SHORT)
                    , c(EZ_MAX_SHORT, EZ_MIN_SHORT, EZ_MAX_SHORT)
                    , d(EZ_MIN_SHORT, EZ_MIN_SHORT, EZ_MAX_SHORT)
                    , e(EZ_MIN_SHORT, EZ_MAX_SHORT, EZ_MIN_SHORT)
                    , f(EZ_MAX_SHORT, EZ_MAX_SHORT, EZ_MIN_SHORT)
                    , g(EZ_MAX_SHORT, EZ_MAX_SHORT, EZ_MAX_SHORT)
                    , h(EZ_MIN_SHORT, EZ_MAX_SHORT, EZ_MAX_SHORT);

        if(rotate_view)
        {
            a = rotate_point(a, AIL->view_angle);
            b = rotate_point(b, AIL->view_angle);
            c = rotate_point(c, AIL->view_angle);
            d = rotate_point(d, AIL->view_angle);
            e = rotate_point(e, AIL->view_angle);
            f = rotate_point(f, AIL->view_angle);
            g = rotate_point(g, AIL->view_angle);
            h = rotate_point(h, AIL->view_angle);
        }
        if(offset_view)
        {
            a += AIL->view_offset;
            b += AIL->view_offset;
            c += AIL->view_offset;
            d += AIL->view_offset;
            e += AIL->view_offset;
            f += AIL->view_offset;
            g += AIL->view_offset;
            h += AIL->view_offset;
        }
        if(scale_view)
        {
            a *= AIL->view_scale;
            b *= AIL->view_scale;
            c *= AIL->view_scale;
            d *= AIL->view_scale;
            e *= AIL->view_scale;
            f *= AIL->view_scale;
            g *= AIL->view_scale;
            h *= AIL->view_scale;
        }

        a *= scale;
        b *= scale;
        c *= scale;
        d *= scale;
        e *= scale;
        f *= scale;
        g *= scale;
        h *= scale;

        bmp_put_line_fast(&BMP,             (int)a.x + bmp_center
                              , BMP.yres - ((int)a.y + bmp_center)
                              ,             (int)b.x + bmp_center
                              , BMP.yres - ((int)b.y + bmp_center)
                              , BMP.gray
                         );
        bmp_put_line_fast(&BMP,             (int)b.x + bmp_center
                              , BMP.yres - ((int)b.y + bmp_center)
                              ,             (int)c.x + bmp_center
                              , BMP.yres - ((int)c.y + bmp_center)
                              , BMP.gray
                         );
        bmp_put_line_fast(&BMP,             (int)c.x + bmp_center
                              , BMP.yres - ((int)c.y + bmp_center)
                              ,             (int)d.x + bmp_center
                              , BMP.yres - ((int)d.y + bmp_center)
                              , BMP.gray
                         );
        bmp_put_line_fast(&BMP,             (int)d.x + bmp_center
                              , BMP.yres - ((int)d.y + bmp_center)
                              ,             (int)a.x + bmp_center
                              , BMP.yres - ((int)a.y + bmp_center)
                              , BMP.gray
                         );
        bmp_put_line_fast(&BMP,             (int)a.x + bmp_center
                              , BMP.yres - ((int)a.y + bmp_center)
                              ,             (int)e.x + bmp_center
                              , BMP.yres - ((int)e.y + bmp_center)
                              , BMP.gray
                         );
        bmp_put_line_fast(&BMP,             (int)b.x + bmp_center
                              , BMP.yres - ((int)b.y + bmp_center)
                              ,             (int)f.x + bmp_center
                              , BMP.yres - ((int)f.y + bmp_center)
                              , BMP.gray
                         );
        bmp_put_line_fast(&BMP,             (int)c.x + bmp_center
                              , BMP.yres - ((int)c.y + bmp_center)
                              ,             (int)g.x + bmp_center
                              , BMP.yres - ((int)g.y + bmp_center)
                              , BMP.gray
                         );
        bmp_put_line_fast(&BMP,             (int)d.x + bmp_center
                              , BMP.yres - ((int)d.y + bmp_center)
                              ,             (int)h.x + bmp_center
                              , BMP.yres - ((int)h.y + bmp_center)
                              , BMP.gray
                         );
        bmp_put_line_fast(&BMP,             (int)e.x + bmp_center
                              , BMP.yres - ((int)e.y + bmp_center)
                              ,             (int)f.x + bmp_center
                              , BMP.yres - ((int)f.y + bmp_center)
                              , BMP.gray
                         );
        bmp_put_line_fast(&BMP,             (int)f.x + bmp_center
                              , BMP.yres - ((int)f.y + bmp_center)
                              ,             (int)g.x + bmp_center
                              , BMP.yres - ((int)g.y + bmp_center)
                              , BMP.gray
                         );
        bmp_put_line_fast(&BMP,             (int)g.x + bmp_center
                              , BMP.yres - ((int)g.y + bmp_center)
                              ,             (int)h.x + bmp_center
                              , BMP.yres - ((int)h.y + bmp_center)
                              , BMP.gray
                         );
        bmp_put_line_fast(&BMP,             (int)h.x + bmp_center
                              , BMP.yres - ((int)h.y + bmp_center)
                              ,             (int)e.x + bmp_center
                              , BMP.yres - ((int)e.y + bmp_center)
                              , BMP.gray
                         );
    }
    //------------------------------------------------------------------------
    if(AIL->show_fulcrum)
    {
        ez_3D_float   x1(-5000, 0, 0)
                    , x2( 5000, 0, 0)
                    , y1(0, -5000, 0)
                    , y2(0,  5000, 0)
                    , z1(0, 0, -5000)
                    , z2(0, 0,  5000);

        x1 = x1 += AIL->fulcrum;
        x2 = x2 += AIL->fulcrum;
        y1 = y1 += AIL->fulcrum;
        y2 = y2 += AIL->fulcrum;
        z1 = z1 += AIL->fulcrum;
        z2 = z2 += AIL->fulcrum;

        if(rotate_view)
        {
            x1 = rotate_point(x1, AIL->view_angle);
            x2 = rotate_point(x2, AIL->view_angle);
            y1 = rotate_point(y1, AIL->view_angle);
            y2 = rotate_point(y2, AIL->view_angle);
            z1 = rotate_point(z1, AIL->view_angle);
            z2 = rotate_point(z2, AIL->view_angle);
        }
        if(offset_view)
        {
            x1 += AIL->view_offset;
            x2 += AIL->view_offset;
            y1 += AIL->view_offset;
            y2 += AIL->view_offset;
        }
        if(scale_view)
        {
            x1 *= AIL->view_scale;
            x2 *= AIL->view_scale;
            y1 *= AIL->view_scale;
            y2 *= AIL->view_scale;
        }
        bmp_put_line_fast(&BMP,             (int)(x1.x * scale) + bmp_center
                              , BMP.yres - ((int)(x1.y * scale) + bmp_center)
                              ,             (int)(x2.x * scale) + bmp_center
                              , BMP.yres - ((int)(x2.y * scale) + bmp_center)
                              , BMP.make_rgb(255,0,0)
                         );
        bmp_put_line_fast(&BMP,             (int)(y1.x * scale) + bmp_center
                              , BMP.yres - ((int)(y1.y * scale) + bmp_center)
                              ,             (int)(y2.x * scale) + bmp_center
                              , BMP.yres - ((int)(y2.y * scale) + bmp_center)
                              , BMP.make_rgb(0,255,0)
                         );
        bmp_put_line_fast(&BMP,             (int)(z1.x * scale) + bmp_center
                              , BMP.yres - ((int)(z1.y * scale) + bmp_center)
                              ,             (int)(z2.x * scale) + bmp_center
                              , BMP.yres - ((int)(z2.y * scale) + bmp_center)
                              , BMP.make_rgb(0,0,255)
                         );
    }
    //------------------------------------------------------------------------
    if(AIL->show_palettes)
        AIL->render_palette_in_ezfb_bitmap();
    //------------------------------------------------------------------------
    if(AIL->show_stats)
    {
        if(format == EZ_UAIL_2D_VECTOR)
        {
            if(is_selected)
                bmp_printf(   &BMP
                            , x_off
                            , y_off
                            , BMP.make_rgb(255,255,0)
                            , BMP.make_rgb(128,0,0)
                            , 0
                            , AIL->font_size_factor
                            , "2D frame %6d"
                            , AIL->frame_index
                          );
            else
                bmp_printf(   &BMP
                            , x_off
                            , y_off
                            , BMP.make_rgb(255,255,0)
                            , bmp_bg_color
                            , 0
                            , AIL->font_size_factor
                            , "2D frame %6d"
                            , AIL->frame_index
                          );
        }
        else
        {
            //----------------------------------------------------------------
            if(is_selected)
                bmp_printf(   &BMP
                            , x_off
                            , y_off
                            , BMP.make_rgb(255,255,0)
                            , BMP.make_rgb(128,0,0)
                            , 0
                            , AIL->font_size_factor
                            , "3D frame %6d"
                            , AIL->frame_index
                          );
            else
                bmp_printf(   &BMP
                            , x_off
                            , y_off
                            , BMP.make_rgb(255,255,0)
                            , bmp_bg_color
                            , 0
                            , AIL->font_size_factor
                            , "3D frame %6d"
                            , AIL->frame_index
                          );
        }
        //--------------------------------------------------------------------
        y_off += y_line;
        if(is_unique)
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.make_rgb(0,255,255)
                        , bmp_bg_color
                        , 0
                        , AIL->font_size_factor
                        , "vertices %6d"
                        , quantity
                      );
        else
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.make_rgb(0,255,255)
                        , BMP.make_rgb(0,128,0)
                        , 0
                        , AIL->font_size_factor
                        , "vertices %6d"
                        , quantity
                      );
        //--------------------------------------------------------------------
        y_off += y_line;
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , BMP.make_rgb(255,0,255)
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "segments %6d"
                    , number_of_segments()
                  );
    }
    //------------------------------------------------------------------------
    // ACTUAL VECTOR ART !!!
    //------------------------------------------------------------------------
    if(quantity > 1)
    {
        //--------------------------------------------------------------------
        if(AIL->show_blanking)
        {
            for(i = 1; i < quantity; i++)
            {
                if(at(i).is_blank())
                {
                    if(rotate_view)
                    {
                        _0 = rotate_point(at(i - 1).bit_masked_position(), AIL->view_angle);
                        _1 = rotate_point(at(i    ).bit_masked_position(), AIL->view_angle);
                    }
                    else
                    {
                        _0 = at(i - 1).bit_masked_position();
                        _1 = at(i    ).bit_masked_position();
                    }
                    if(offset_view)
                    {
                        _0 += AIL->view_offset;
                        _1 += AIL->view_offset;
                    }
                    if(scale_view)
                    {
                        _0 *= AIL->view_scale;
                        _1 *= AIL->view_scale;
                    }
                    xx = (int)(_1.x * scale) + bmp_center;
                    yy = (int)(_1.y * scale) + bmp_center;

                    bmp_put_line_fast(&BMP,             (int)(_0.x * scale) + bmp_center
                                          , BMP.yres - ((int)(_0.y * scale) + bmp_center)
                                          ,             xx
                                          , BMP.yres - (yy)
                                          , BMP.gray
                                     );
                    if(AIL->show_points)
                        bmp_put_vertex(&BMP, xx - 1
                                           , BMP.yres - yy
                                           , BMP.gray
                                      );
                    if(AIL->show_cursors && is_index_selected(i))
                        bmp_put_select(&BMP, xx
                                           , BMP.yres - yy
                                           , BMP.gray
                                      );
                }
            }
        } // end if(AIL->show_blanking)
        //--------------------------------------------------------------------
        for(i = 1; i < quantity; i++)
        {            
            if(at(i).is_black())
            {
                if(rotate_view)
                {
                    _0 = rotate_point(at(i - 1).bit_masked_position(), AIL->view_angle);
                    _1 = rotate_point(at(i    ).bit_masked_position(), AIL->view_angle);
                }
                else
                {
                    _0 = at(i - 1).bit_masked_position();
                    _1 = at(i    ).bit_masked_position();
                }
                if(offset_view)
                {
                    _0 += AIL->view_offset;
                    _1 += AIL->view_offset;
                }
                if(scale_view)
                {
                    _0 *= AIL->view_scale;
                    _1 *= AIL->view_scale;
                }
                xx = (int)(_1.x * scale) + bmp_center;
                yy = (int)(_1.y * scale) + bmp_center;
                //------------------------------------------------------------
                bmp_put_line_fast(&BMP,             (int)(_0.x * scale) + bmp_center
                                      , BMP.yres - ((int)(_0.y * scale) + bmp_center)
                                      ,             xx
                                      , BMP.yres - (yy)
                                      , 0x00000000
                                 );
                if(AIL->show_points)
                    bmp_put_vertex(&BMP, xx
                                       , BMP.yres - yy
                                       , 0x00000000
                                  );
                //------------------------------------------------------------
                if(AIL->show_cursors && is_index_selected(i))
                    bmp_put_select(&BMP, xx
                                       , BMP.yres - yy
                                       , BMP.white
                                  );
            }
        }
        //--------------------------------------------------------------------
        for(i = 1; i < quantity; i++)
        {            
            if(at(i).is_color())
            {
                if(rotate_view)
                {
                    _0 = rotate_point(at(i - 1).bit_masked_position(), AIL->view_angle);
                    _1 = rotate_point(at(i    ).bit_masked_position(), AIL->view_angle);
                }
                else
                {
                    _0 = at(i - 1).bit_masked_position();
                    _1 = at(i    ).bit_masked_position();
                }
                if(offset_view)
                {
                    _0 += AIL->view_offset;
                    _1 += AIL->view_offset;
                }
                if(scale_view)
                {
                    _0 *= AIL->view_scale;
                    _1 *= AIL->view_scale;
                }

                xx = (int)(_1.x * scale) + bmp_center;
                yy = (int)(_1.y * scale) + bmp_center;
                bit_masked = at(i).bit_masked_color();
                //------------------------------------------------------------
                bmp_put_line_fast(&BMP,             (int)(_0.x * scale) + bmp_center
                                      , BMP.yres - ((int)(_0.y * scale) + bmp_center)
                                      ,             xx
                                      , BMP.yres - (yy)
                                      , BMP.make_rgb(bit_masked.r, bit_masked.g, bit_masked.b)
                                 );
                if(AIL->show_points)
                    bmp_put_vertex(&BMP, xx
                                       , BMP.yres - yy
                                       , BMP.make_rgb(bit_masked.r, bit_masked.g, bit_masked.b)
                                  );
                //------------------------------------------------------------
                if(AIL->show_cursors && is_index_selected(i))
                    bmp_put_select(&BMP, xx
                                       , BMP.yres - yy
                                       ,   at(i).is_black()
                                         ? BMP.white
                                         : BMP.make_rgb(bit_masked.r, bit_masked.g, bit_masked.b)
                                  );
            }
        }
        //--------------------------------------------------------------------
        // END ACTUAL VECTOR ART !!!
        //--------------------------------------------------------------------
        if(AIL->show_cursors)
        {
            bit_masked = at(egg).bit_masked_color();
            select_start_color =   at(egg).is_blank()
                                 ? BMP.gray
                                 :   at(egg).is_black()
                                   ? BMP.white
                                   : BMP.make_rgb(bit_masked.r, bit_masked.g, bit_masked.b);
            //----------------------------------------------------------------
            bit_masked = at(spider).bit_masked_color();
            select_end_color   =   at(spider).is_blank()
                                 ? BMP.gray
                                 :   at(spider).is_black()
                                   ? BMP.white
                                   : BMP.make_rgb(bit_masked.r, bit_masked.g, bit_masked.b);
            //----------------------------------------------------------------
            if(rotate_view)
            {
                _0 = rotate_point(at(egg   ).bit_masked_position(), AIL->view_angle);
                _1 = rotate_point(at(spider).bit_masked_position(), AIL->view_angle);
            }
            else
            {
                _0 = at(egg   ).bit_masked_position();
                _1 = at(spider).bit_masked_position();
            }
            if(offset_view)
            {
                _0 += AIL->view_offset;
                _1 += AIL->view_offset;
            }
            if(scale_view)
            {
                _0 *= AIL->view_scale;
                _1 *= AIL->view_scale;
            }
            //----------------------------------------------------------------
            // start point select cursor
            bmp_put_rectangle(&BMP,             (int)((_0.x - 1000) * scale) + bmp_center
                                  , BMP.yres - ((int)((_0.y - 1000) * scale) + bmp_center)
                                  ,             (int)((_0.x + 1000) * scale) + bmp_center
                                  , BMP.yres - ((int)((_0.y + 1000) * scale) + bmp_center)
                                  , select_start_color
                                  , 0
                             );
            //----------------------------------------------------------------
            // end point select cursor
            bmp_put_line_fast (&BMP,             (int)((_1.x       ) * scale) + bmp_center
                                   , BMP.yres - ((int)((_1.y - 1000) * scale) + bmp_center)
                                   ,             (int)((_1.x       ) * scale) + bmp_center
                                   , BMP.yres - ((int)((_1.y + 1000) * scale) + bmp_center)
                                   , select_end_color
                              );
            bmp_put_line_fast (&BMP,             (int)((_1.x + 1000) * scale) + bmp_center
                                   , BMP.yres - ((int)((_1.y       ) * scale) + bmp_center)
                                   ,             (int)((_1.x - 1000) * scale) + bmp_center
                                   , BMP.yres - ((int)((_1.y       ) * scale) + bmp_center)
                                   , select_end_color
                              );
            bmp_put_line_fast (&BMP,             (int)((_1.x - 1000) * scale) + bmp_center
                                   , BMP.yres - ((int)((_1.y - 1000) * scale) + bmp_center)
                                   ,             (int)((_1.x + 1000) * scale) + bmp_center
                                   , BMP.yres - ((int)((_1.y + 1000) * scale) + bmp_center)
                                   , select_end_color
                              );
            bmp_put_line_fast (&BMP,             (int)((_1.x + 1000) * scale) + bmp_center
                                   , BMP.yres - ((int)((_1.y - 1000) * scale) + bmp_center)
                                   ,             (int)((_1.x - 1000) * scale) + bmp_center
                                   , BMP.yres - ((int)((_1.y + 1000) * scale) + bmp_center)
                                   , select_end_color
                              );
            //----------------------------------------------------------------
            // icons
            y_off += (2 * y_line); // color vertex segment for start cursor            
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.make_rgb(255,0,255)
                        , bmp_bg_color
                        , 0
                        , AIL->font_size_factor
                        , "           %4d"
                        , segment_index_of_point(egg)
                      );
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.make_rgb(0,255,255)
                        , bmp_bg_color
                        , 0
                        , AIL->font_size_factor
                        , "      %4d"
                        , egg
                      );
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.white
                        , bmp_bg_color
                        , 0
                        , AIL->font_size_factor
                        , "  %3d"
                        , at(egg).c
                      );
            //----------------------------------------------------------------
            y_off += y_line; // RGB values for start cursor
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.make_rgb(0,0,255)
                        , bmp_bg_color
                        , 0
                        , AIL->font_size_factor
                        , "            %3d"
                        , at(egg).b
                      );
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.make_rgb(0,255,0)
                        , bmp_bg_color
                        , 0
                        , AIL->font_size_factor
                        , "        %3d"
                        , at(egg).g
                      );
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.make_rgb(255,0,0)
                        , bmp_bg_color
                        , 0
                        , AIL->font_size_factor
                        , "    %3d"
                        , at(egg).r
                      );
            //----------------------------------------------------------------
            y_off += y_line; // color vertex segment for end cursor
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.make_rgb(255,0,255)
                        , bmp_bg_color
                        , 0
                        , AIL->font_size_factor
                        , "           %4d"
                        , segment_index_of_point(spider)
                      );
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.make_rgb(0,255,255)
                        , bmp_bg_color
                        , 0
                        , AIL->font_size_factor
                        , "      %4d"
                        , spider
                      );
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.white
                        , bmp_bg_color
                        , 0
                        , AIL->font_size_factor
                        , "  %3d"
                        , at(spider).c
                      );
            //----------------------------------------------------------------
            y_off += y_line; // RGB values for end cursor
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.make_rgb(0,0,255)
                        , bmp_bg_color
                        , 0
                        , AIL->font_size_factor
                        , "            %3d"
                        , at(spider).b
                      );
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.make_rgb(0,255,0)
                        , bmp_bg_color
                        , 0
                        , AIL->font_size_factor
                        , "        %3d"
                        , at(spider).g
                      );
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.make_rgb(255,0,0)
                        , bmp_bg_color
                        , 0
                        , AIL->font_size_factor
                        , "    %3d"
                        , at(spider).r
                      );
            //-----------------------------------------------------------------
            y_off -= (3 * y_line);
            bmp_put_rectangle(&BMP, x_off
                                  , y_off
                                  , x_off + (int)(2000 * scale)
                                  , y_off + (int)(2000 * scale)
                                  , select_start_color
                                  , 0
                             ); // start cursor icon
            //----------------------------------------------------------------
            y_off += (2 * y_line);
            bmp_put_line_fast     (&BMP, x_off + (int)(1000 * scale)
                                       , y_off
                                       , x_off + (int)(1000 * scale)
                                       , y_off + (int)(2000 * scale)
                                       , select_end_color
                                  );
            bmp_put_line_fast     (&BMP, x_off + (int)(2000 * scale)
                                       , y_off + (int)(1000 * scale)
                                       , x_off
                                       , y_off + (int)(1000 * scale)
                                       , select_end_color
                                  );
            bmp_put_line_fast     (&BMP, x_off
                                       , y_off
                                       , x_off + (int)(2000 * scale)
                                       , y_off + (int)(2000 * scale)
                                       , select_end_color
                                  );
            bmp_put_line_fast     (&BMP, x_off + (int)(2000 * scale)
                                       , y_off
                                       , x_off
                                       , y_off + (int)(2000 * scale)
                                       , select_end_color
                                  ); // end cursor icon
            //----------------------------------------------------------------
        } // end if(AIL->show_cursors)
        //--------------------------------------------------------------------
    } // end if(quantity > 1)
    //------------------------------------------------------------------------
    if(    AIL->show_ui_message
        && ui_message.size()
      )
    {
        y_off = BMP.yres - y_line;
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , BMP.make_rgb(255,255,0)
                    , BMP.make_rgb(63,0,63)
                    , 0
                    , AIL->font_size_factor
                    , "%s"
                    , ui_message.c_str()
                  );
    }
    return;
}

//############################################################################
void ez_uail_frame::save_as_wave(string file)
{
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".wav"
      )
        file += ".wav";
    ofstream out((EZ_UAIL_WAVE_SHARE + file).c_str(), ios::out | ios::binary);
    //------------------------------------------------------------------------
    if(out.is_open())
    {
        if(quantity > 1)
        {
            int              i,
                             frame_scan,
                             blank_dwell = 0,
                             total_frame_scans;
                             
            ez_uail_point    point;
            
            ez_uail_segment  intro,
                             bridge,
                             coda;
                             
            ez_wave_header   header(AIL->sample_rate,
                                      EZ_WAVE_POSITIVE
//                                    | EZ_WAVE_OFFSETS
                                    | EZ_WAVE_END_OF_FRAME
                                    | EZ_WAVE_UNIQUE_FRAME
                                    | EZ_WAVE_SIGNAL_MATRIX,
                                    6
                                   );

            ez_uail_frame    copy(*this);
            //----------------------------------------------------------------
//            header.offset    [0] = 0; // default values
//            header.offset    [1] = 0;
//            header.offset    [2] = 0;
//            header.offset    [3] = 0;
//            header.offset    [4] = 0;
//            header.offset    [5] = 0;

            header.signal_id [0] = EZ_SIGNAL_X_POSITION  ; // default values
            header.signal_id [1] = EZ_SIGNAL_Y_POSITION  ;
            header.signal_id [2] = EZ_SIGNAL_RED_ANALOG  ;
            header.signal_id [3] = EZ_SIGNAL_GREEN_ANALOG;
            header.signal_id [4] = EZ_SIGNAL_BLUE_ANALOG ;
            header.signal_id [5] = EZ_SIGNAL_Z_POSITION  ;
            
            header.LSB_tag   [0] = EZ_LSB_NOT_USED    ; // default values
            header.LSB_tag   [1] = EZ_LSB_NOT_USED    ;
            header.LSB_tag   [2] = EZ_LSB_END_OF_FRAME;
            header.LSB_tag   [3] = EZ_LSB_UNIQUE_FRAME;
            header.LSB_tag   [4] = EZ_LSB_NOT_USED    ;
            header.LSB_tag   [5] = EZ_LSB_NOT_USED    ;
                
            if(    AIL->signal_bit_mask[0] != 0
                || AIL->signal_bit_mask[1] != 0
                || AIL->signal_bit_mask[2]  > 7
                || AIL->signal_bit_mask[3]  > 7
                || AIL->signal_bit_mask[4]  > 7
                || AIL->signal_bit_mask[5] != 0
                || AIL->signal_bit_mask[6] != 0
                || AIL->signal_bit_mask[7] != 0
              )
            {
                header.resolution[0] = 16 - AIL->signal_bit_mask[0];
                header.resolution[1] = 16 - AIL->signal_bit_mask[1];
                header.resolution[2] = 16 - AIL->signal_bit_mask[2];
                header.resolution[3] = 16 - AIL->signal_bit_mask[3];
                header.resolution[4] = 16 - AIL->signal_bit_mask[4];
                header.resolution[5] = 16 - AIL->signal_bit_mask[5];
                header.resolution[6] = 16 - AIL->signal_bit_mask[6];
                header.resolution[7] = 16 - AIL->signal_bit_mask[7];
                header.ez_wave_mode |= EZ_SIGNAL_BIT_RESOLUTION;                
            }        
            //----------------------------------------------------------------
            copy.rotate_around_origin();
            
            if(copy.is_2D() || AIL->auto_flatten_z)
            {    
                header.signal_id [5] = AIL->signal_if_not_z;
                header.resolution[5] = 8;
                copy.flatten_z();
            }

            if(AIL->auto_optimize)
            {
                header.ez_wave_mode |= EZ_WAVE_OPTIMIZED;
                header.parms.lit_dwell_overhang       = AIL->lit_dwell_overhang      ;
                header.parms.lit_distance_delta_max   = AIL->lit_distance_delta_max  ;
                header.parms.blank_distance_delta_max = AIL->blank_distance_delta_max;
                header.parms.longest_dwell_microsec   = AIL->longest_dwell_microsec  ;
                header.parms.insignificant_distance   = AIL->insignificant_distance  ;
                header.parms.zero_angle_tolerance     = AIL->zero_angle_tolerance    ;
                header.parms.frames_per_second        = AIL->frames_per_second       ;
                copy.optimize();
            }
            //----------------------------------------------------------------
            total_frame_scans = (int)ceil((   AIL->sample_rate
                                            * AIL->still_frame_duration
                                          ) / copy.quantity
                                         );                                         
            //----------------------------------------------------------------
            if(AIL->auto_optimize)
            {
                blank_dwell = (int)(   AIL->sample_rate
                                     / AIL->longest_dwell_microsec
                                   );
                //------------------------------------------------------------
                for(i = 0; i < blank_dwell; i++)
                    intro.push_back(ez_uail_point()); // the origin
                //------------------------------------------------------------
                point = copy.first_lit_anchor();
                point.blank();
                intro += ez_uail_segment(intro[0], point); // line from origin to first point
                //------------------------------------------------------------
                bridge = ez_uail_segment(copy.last_lit_point(), point); // line from last point to first for repete
                bridge.blank_all_points();
                //------------------------------------------------------------
                coda = ez_uail_segment(point, intro[0]); // line from last point to origin
                for(i = 0; i < blank_dwell; i++)
                    coda.push_back(ez_uail_point()); // the origin
            } // end if(AIL->auto_optimize)
            //----------------------------------------------------------------
            if(AIL->invert_wave_output)
                header.ez_wave_mode &= ~EZ_WAVE_POSITIVE;
            //----------------------------------------------------------------
            header.to_ofstream_wave(out); // take up the space!
            //----------------------------------------------------------------
            if(AIL->auto_optimize)
                intro.to_ofstream_wave(out, header, false, false);
            //----------------------------------------------------------------
            ((ez_uail_segment)copy).to_ofstream_wave(out, header, true, is_unique); // first frame unique
            //----------------------------------------------------------------
            for(frame_scan = 1; frame_scan < total_frame_scans; frame_scan++)
            {
                bridge.to_ofstream_wave(out, header, false, false);
                ((ez_uail_segment)copy).to_ofstream_wave(out, header, true, false); // the rest are not
            }
            //----------------------------------------------------------------
            if(AIL->auto_optimize)
                coda.to_ofstream_wave(out, header, false, false);
            //----------------------------------------------------------------
            header.to_ofstream_wave(out);
            //----------------------------------------------------------------
        }
        out.close();
    }
    return;
}

//############################################################################
void ez_uail_frame::to_ofstream_wave(ofstream& out, ez_wave_header& header)
{
    if(quantity > 1)
    {
        int frame_scan,
            total_frame_scans = 1;
        //--------------------------------------------------------------------
        ez_uail_segment bridge(last_lit_point(), first_lit_anchor()); // line from last point to first for repete
        bridge.blank_all_points();
        //--------------------------------------------------------------------
        if(AIL->frames_per_second > 0.00)
        {
            //----------------------------------------------------------------
            if(AIL->auto_optimize)
                total_frame_scans = (int)ceil(   (    AIL->sample_rate
                                                    / (   quantity
                                                        + bridge.number_of_points()
                                                      )
                                                  )
                                                / AIL->frames_per_second
                                             );
            //----------------------------------------------------------------
            else
                total_frame_scans = (int)ceil(   (    AIL->sample_rate 
                                                    / quantity
                                                  )
                                                / AIL->frames_per_second
                                             );
        } // end if(AIL->frames_per_second > 0.00)
        //--------------------------------------------------------------------
        ez_uail_segment::to_ofstream_wave(out, header, true, true); // first pass through this frame
        //--------------------------------------------------------------------
        if(total_frame_scans > 1)
            for(frame_scan = 1; frame_scan < total_frame_scans; frame_scan++)
            {
                bridge.to_ofstream_wave(out, header, false, false);
                ez_uail_segment::to_ofstream_wave(out, header, true, false);
            }
        //--------------------------------------------------------------------
        transition_to_next_frame.to_ofstream_wave(out, header, false, false);
    }
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void ez_uail_frame::to_ofstream_code(ofstream& out, const char* indent) const
{
    out << indent
        << "frame.format = "
        <<  (int)format
        <<  ";\n"
        <<  indent
        << "frame.palette_index = "
        <<  palette_index
        <<  ";\n";

    for(int i = 0; i < quantity; i++)
        out << indent
            <<  "frame.push_back(ez_uail_point("
            << setw(6)
            <<  at(i).x
            <<  ", "
            << setw(6)
            <<  at(i).y
            <<  ", "
            << setw(6)
            <<  at(i).z
            <<  ", "
            << setw(3)
            <<  (int)at(i).r
            <<  ", "
            << setw(3)
            <<  (int)at(i).g
            <<  ", "
            << setw(3)
            <<  (int)at(i).b
            <<  ", "
            << setw(3)
            <<  (int)at(i).k
            <<  ", "
            << setw(3)
            <<  (int)at(i).c
            <<  "));\n";

    out << indent << "frame.quantity = (u_short)frame.quantity;\n";
    return;
}

//############################################################################
void ez_uail_frame::save_as_code(string object_name) const
{
    ofstream out((EZ_UAIL_CPP_SHARE + object_name + ".cpp").c_str(), ios::out);
    if(out.is_open())
    {
        out << "//############################################################################\n"
               "// ~  This file is the product of LaserBoy!  ~\n"
               "//\n"
               "// by James Lehman\n"
               "// james@akrobiz.com\n"
               "//\n"
               "//############################################################################\n\n\n\n"
               "//############################################################################\n"

               "ez_uail_frame "
            << object_name
            << "()\n"
               "{\n"
               "    ez_uail_frame frame;\n";

        to_ofstream_code(out, "    ");

        out << "    return frame;\n"
               "}\n\n"
               "//############################################################################\n"
               "//////////////////////////////////////////////////////////////////////////////\n"
               "//############################################################################\n\n\n";

        out.close();
    }
    return;
}

//############################################################################
void ez_uail_frame::save_as_dxf(string file) const
{
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".dxf"
      )
        file += ".dxf";
    ofstream out((EZ_UAIL_DXF_SHARE + file).c_str(), ios::out);
    if(out.is_open())
    {
        int                segment_vector_index,
                           point_index;
        ez_uail_frame_set  segments = find_segments();
        //--------------------------------------------------------------------
        segments.minimize();
        //--------------------------------------------------------------------
        out << " 0\n"
                "SECTION\n"
                " 2\n"
                "ENTITIES\n";
        //--------------------------------------------------------------------
        for(segment_vector_index = 0; segment_vector_index < number_of_segments(); segment_vector_index++)
        {
//            if(segments[segment_vector_index].number_of_lit_points() == 1)
//            {
//                out << " 0\n"
//                       "POINT\n"
//                       " 62\n  ";
//                out << AIL->
//                       palette_picker(EZ_DXF)
//                       .best_match(segments[segment_vector_index].first_lit_point());
//                out << "\n 10\n";
//                out << segments[segment_vector_index].first_lit_anchor().x;
//                out << "\n 20\n";
//                out << segments[segment_vector_index].first_lit_anchor().y;
//                out << "\n 30\n";
//                out << segments[segment_vector_index].first_lit_anchor().z;
//                out << "\n";
//            }
            //----------------------------------------------------------------
            if(segments[segment_vector_index].number_of_lit_points() == 1)
            {
                out << " 0\n"
                       "LINE\n"
                       " 62\n  ";
                out << AIL->
                       palette_picker(EZ_DXF)
                       .best_match(segments[segment_vector_index].first_lit_point());
                out << "\n 10\n";
                out << segments[segment_vector_index].first_lit_anchor().x;
                out << "\n 20\n";
                out << segments[segment_vector_index].first_lit_anchor().y;
                out << "\n 30\n";
                out << segments[segment_vector_index].first_lit_anchor().z;
                out << "\n 11\n";
                out << segments[segment_vector_index].first_lit_point().x;
                out << "\n 21\n";
                out << segments[segment_vector_index].first_lit_point().y;
                out << "\n 31\n";
                out << segments[segment_vector_index].first_lit_point().z;
                out << "\n";
            }
            //----------------------------------------------------------------
            else
            {
                out << " 0\n"
                        "POLYLINE\n"
                        " 70\n";
                out << segments[segment_vector_index].is_closed_polygon();
                out << "\n 10\n"
                       "0.0\n"
                       " 20\n"
                       "0.0\n"
                       " 30\n"
                       "0.0\n"
                       //---------
                       " 0\n"
                       "VERTEX\n"
                       " 62\n";
                out << AIL->
                       palette_picker(EZ_DXF)
                       .best_match(segments[segment_vector_index].at(1));
                out << "\n 10\n";
                out << segments[segment_vector_index].first_lit_anchor().x;
                out << "\n 20\n";
                out << segments[segment_vector_index].first_lit_anchor().y;
                out << "\n 30\n";
                out << segments[segment_vector_index].first_lit_anchor().z;
                out << "\n";

                for(point_index = 1; point_index < segments[segment_vector_index].number_of_points() - 1; point_index++)
                {
                    out << " 0\n"
                        << "VERTEX\n"
                        << " 62\n";
                    out << AIL->
                           palette_picker(EZ_DXF)
                           .best_match(segments[segment_vector_index].at(point_index + 1));
                    out << "\n 10\n";
                    out << segments[segment_vector_index].at(point_index).x;
                    out << "\n 20\n";
                    out << segments[segment_vector_index].at(point_index).y;
                    out << "\n 30\n";
                    out << segments[segment_vector_index].at(point_index).z;
                    out << "\n";
                }
                out << " 0\n"
                       "VERTEX\n"
                       " 62\n"
                       "0\n"
                       " 10\n";
                out << segments[segment_vector_index].at(point_index).x;
                out << "\n 20\n";
                out << segments[segment_vector_index].at(point_index).y;
                out << "\n 30\n";
                out << segments[segment_vector_index].at(point_index).z;
                out << "\n";
            }
        }
        //--------------------------------------------------------------------
        out << " 0\n"
               "ENDSEC\n"
               " 0\n"
               "EOF\n";
        //--------------------------------------------------------------------
        out.close();
    }
    return;
}

//############################################################################
void ez_uail_frame::save_as_bmp(string file) const
{
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".bmp"
      )
        file += ".bmp";
    //------------------------------------------------------------------------
    if(quantity > 1)
    {
        char   file_name[80];
        struct ezfb_bitmap bmp = {0};

        if(palette_index == EZ_TRUE_COLOR)
            bmp_init(&bmp, AIL->output_bmp_size, AIL->output_bmp_size, 24);
        else
        {
            bmp_init(&bmp, AIL->output_bmp_size, AIL->output_bmp_size,  8);
            AIL->palette_picker(palette_index).to_ezfb_bmp_palette(&bmp);
        }
        bmp_clear(&bmp, bmp.black);

        u_short         i;
        ez_3D_float    _0,
                       _1;
        int            bmp_center  = bmp.yres / 2;
        float          scale       = bmp.yres / 65536.0;
        //--------------------------------------------------------------------
        for(i = 1; i < quantity; i++)
        {
            if(at(i).is_color())
            {
                _0 = at(i - 1);
                _1 = at(i    );
                //------------------------------------------------------------

                if(palette_index == EZ_TRUE_COLOR)
                    bmp_put_line_fast(&bmp,  (int)(_0.x * scale) + bmp_center
                                          , ((int)(_0.y * scale) + bmp_center)
                                          ,  (int)(_1.x * scale) + bmp_center
                                          , ((int)(_1.y * scale) + bmp_center)
                                          , BMP.make_rgb(at(i).r, at(i).g, at(i).b)
                                     );
                else
                    bmp_put_line_fast(&bmp,  (int)(_0.x * scale) + bmp_center
                                          , ((int)(_0.y * scale) + bmp_center)
                                          ,  (int)(_1.x * scale) + bmp_center
                                          , ((int)(_1.y * scale) + bmp_center)
                                          , at(i).c
                                     );
            }
        }
        strcpy(file_name, (EZ_UAIL_BMP_SHARE + file).c_str());
        bmp_save_as(&bmp, file_name);
        bmp_free(&bmp);
    }
    return;
}

//############################################################################
void ez_uail_frame::display_stats() const
{
    int   points       = quantity      ,
          lit          = number_of_lit_points()  ,
          blank        = number_of_blank_points(),
          y_line       = 8 * AIL->font_size_factor,
          x_off        = EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor,
          y_off        = BMP.yres - 19 * y_line; // 19 lines from the bottom        
    u_int bmp_bg_color = BMP.make_rgb(AIL->bg_color.r, AIL->bg_color.g, AIL->bg_color.b);

    ez_3D_short  masked;

    bmp_printf(   &BMP, x_off, y_off, BMP.white, bmp_bg_color, 0, AIL->font_size_factor
                , "Frame stats"
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, AIL->font_size_factor
                , "frame is %s"
                , ((is_selected) ? ("selected") : ("not selected"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, AIL->font_size_factor
                , "frame is %s"
                , ((is_unique) ? ("unique") : ("not unique"))
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, AIL->font_size_factor
                , "segments ....... %d"
                , number_of_segments()
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, AIL->font_size_factor
                , "vertices........ %d"
                , points
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, AIL->font_size_factor
                , "lit vertices.... %d"
                , lit
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, AIL->font_size_factor
                , "blank vertices.. %d"
                , blank
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, AIL->font_size_factor
                , "................ %f%% lit"
                , (lit/(float)points) * 100.0
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, AIL->font_size_factor
                , "................ %f%% blank"
                , (blank/(float)points) * 100.0
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, AIL->font_size_factor
                , "longest vector . %f points"
                , max_distance()
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, AIL->font_size_factor
                , "total length ... %f points"
                , total_distance()
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, AIL->font_size_factor
                , "sharpest angle . %f%c"
                , max_angle() * one_radian
                , 0xf8
              );
    y_off += y_line;
    bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, AIL->font_size_factor
                , "total angle .... %f%c"
                , total_angle() * one_radian
                , 0xf8
              );
    if(AIL->show_cursors)
    {
        masked = at(egg).bit_masked_position();        
        y_off += y_line;
        bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, AIL->font_size_factor
                    , "egg .. %c %5d (X %6d) (Y %6d) (Z %6d)"
                    , ((at(egg).is_lit()) ? ('L') : ('B'))
                    , egg
                    , masked.x
                    , masked.y
                    , masked.z
                  );
        masked = at(spider).bit_masked_position();        
        y_off += y_line;
        bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, AIL->font_size_factor
                    , "spider %c %5d (X %6d) (Y %6d) (Z %6d)"
                    , ((at(spider).is_lit()) ? ('L') : ('B'))                    
                    , spider
                    , masked.x
                    , masked.y
                    , masked.z
                  );
        y_off += y_line;
        bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, AIL->font_size_factor
                    , "select center  (X %6.0f) (Y %6.0f) (Z %6.0f)"
                    , selected_segment().find_center().x
                    , selected_segment().find_center().y
                    , selected_segment().find_center().z
                  );
    }
    if(AIL->show_fulcrum)
    {
        y_off += y_line;
        bmp_printf(   &BMP, x_off, y_off, next_hue(), bmp_bg_color, 0, AIL->font_size_factor
                    , "fulcrum ...... (X %6.0f) (Y %6.0f) (Z %6.0f)"
                    , AIL->fulcrum.x
                    , AIL->fulcrum.y
                    , AIL->fulcrum.z
                  );
    }
    bmp_to_SDL_Surface(&BMP, screen);
    return;
}

//############################################################################






//############################################################################
ez_uail_frame blank_frame()
{
    ez_uail_frame frame(blank_segment(), ez_uail_header());
    frame.quantity = (u_short)frame.quantity;
    frame.is_2D();
    return frame;
}

//############################################################################
ez_uail_frame NULL_frame()
{
    ez_uail_frame frame(NULL_segment(), ez_uail_header());
    frame.quantity = (u_short)frame.quantity;
    frame.is_2D();
    return frame;
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
