//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#ifndef __EZ_WAVE_DEFINITIONS__
#define __EZ_WAVE_DEFINITIONS__

//############################################################################
#include "ez_uail_common.hpp"

//############################################################################
class ez_wave_optimization_parameters
{
public:
    ez_wave_optimization_parameters()
              : lit_dwell_overhang      (-1), // -1 indicates undefined
                lit_distance_delta_max  (-1),
                blank_distance_delta_max(-1),
                longest_dwell_microsec  (-1),
                insignificant_distance  (-1),
                zero_angle_tolerance    (-1),
                frames_per_second       (-1)
                                                {}
    //------------------------------------------------------------------------
   ~ez_wave_optimization_parameters()           {}
    //------------------------------------------------------------------------
    int    lit_dwell_overhang      ;
    float  lit_distance_delta_max  ,
           blank_distance_delta_max,
           longest_dwell_microsec  ,
           insignificant_distance  ,
           zero_angle_tolerance    ,
           frames_per_second       ;
};

//############################################################################
class ez_wave_header
{
public:
    ez_wave_header(int r, u_int m, short c)
                  : num_samples     (0),
                    sample_rate     (r),
                    ez_wave_mode    (m),
                    num_channels    (c),
                    bits_per_sample (16),
                    version         (LASERBOY_WAVE_VERSION),
                    parms           ()
                                                            {}
    //------------------------------------------------------------------------
    ez_wave_header    (const ez_wave_header& header     );
    ez_wave_header    (const ez_wave_header& header, int); // copy 8ch from 6ch
    ez_wave_header    (ifstream& wave_in                );
    //------------------------------------------------------------------------
   ~ez_wave_header    ()                                    {}
    //------------------------------------------------------------------------
    void     to_ofstream_wave (ofstream& out) const;
    void     tell             (string label ) const;
    //------------------------------------------------------------------------
    int      num_samples    ,
             sample_rate    ,
             offset      [8];
    //------------------------------------------------------------------------
    u_char   resolution  [8];
    //------------------------------------------------------------------------
    u_int    ez_wave_mode   ; // up to 32 binary flags
    //------------------------------------------------------------------------
    short    num_channels   ,
             bits_per_sample,
             signal_id   [8],
             LSB_tag     [8];
    //------------------------------------------------------------------------
    string   version        ;
    //------------------------------------------------------------------------
    ez_wave_optimization_parameters parms;
};

//############################################################################
class ez_wave_sample
{
public:
    ez_wave_sample(int n)
        {
            num_channels = n;
            channel = new short[num_channels];
            memset((void*)channel, 0x00, num_channels * sizeof(short));
        }
    //------------------------------------------------------------------------
   ~ez_wave_sample()    {    delete channel;    }
    //------------------------------------------------------------------------
    bool from_ifstream_wave (ifstream& in)
            {
                if(!in.eof())
                {
                    in.read((char*)channel, num_channels * sizeof(signed short));
                    return true;
                }
                return false; // returns false if EOF
            }
    //------------------------------------------------------------------------
    bool from_ifstream_wave_inverted(ifstream& in)
            {
                if(!in.eof())
                {
                    in.read((char*)channel, num_channels * sizeof(signed short));
                    negate();
                    return true;
                }
                return false; // returns false if EOF
            }
    //------------------------------------------------------------------------
    bool from_ifstream_wave_polarity(ifstream& in, int polarity[8])
            {
                if(!in.eof())
                {
                    in.read((char*)channel, num_channels * sizeof(signed short));
                    for(int i = 0; i < num_channels; i++)
                        if(polarity[i] == 1)
                        {
                            if(channel[i] >= EZ_MIN_SHORT)
                                channel[i] = -channel[i];
                            else
                                channel[i] = EZ_MAX_SHORT;
                        }
                    return true;
                }
                return false; // returns false if EOF
            }
    //------------------------------------------------------------------------
    void to_ofstream_wave(ofstream& out, ez_wave_header& header)
            {
                out.write((char*)channel, num_channels * sizeof(signed short));
                header.num_samples++;
                return;
            }
    //------------------------------------------------------------------------
    void to_ofstream_wave_inverted(ofstream& out, ez_wave_header& header)
            {
                negate();
                out.write((char*)channel, num_channels * sizeof(signed short));
                negate();
                header.num_samples++;
                return;
            }
    //------------------------------------------------------------------------
    ez_wave_sample& operator = (const ez_wave_sample& sample)
                    {
                        delete channel;
                        num_channels = sample.num_channels;
                        channel = new short[num_channels];
                        memcpy((void*)channel, (void*)sample.channel, num_channels * sizeof(short));
                        return *this;
                    }
    //------------------------------------------------------------------------
    void negate()
            {
                for(int i = 0; i < num_channels; i++)
                    channel[i] = -channel[i];
                return;
            }
    //------------------------------------------------------------------------
    ez_wave_sample  operator + (const ez_wave_sample& s) // side-by-side as tracks
                    {
                        ez_wave_sample sum(num_channels + s.num_channels);
                        memcpy(   (void*)sum.channel,
                                  (void*)channel,
                                  num_channels * sizeof(short)
                              );                              
                        memcpy(   (void*)(sum.channel + num_channels), // + offset for above
                                  (void*)s.channel,
                                  s.num_channels * sizeof(short)
                              );
                        return sum;
                    }
    //------------------------------------------------------------------------
    signed short   *channel     ;
    int             num_channels;
};

//############################################################################
typedef ez_wave_sample*  ez_wave_sample_pointer;

//############################################################################
#endif

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
